# Optional order type mapping.
order_types = {
    "entry": "limit",
    "exit": "limit",
    "stoploss": "market",
    "stoploss_on_exchange": False
}

# Optional order time in force.
order_time_in_force = {
    "entry": "GTC",
    "exit": "GTC"
}
